""" Transforms involving ECEF: earth-centered, earth-fixed frame """
try:
    from numpy import radians, sin, cos, tan, arctan as atan, hypot, degrees, arctan2 as atan2, sqrt, pi, vectorize
    from .eci import eci2ecef, ecef2eci
except ImportError:
    from math import radians, sin, cos, tan, atan, hypot, degrees, atan2, sqrt, pi

    vectorize = eci2ecef = None
import typing
from datetime import datetime

from .ellipsoid import Ellipsoid
from .utils import sanitize

# py < 3.6 compatible
tau = 2 * pi

__all__ = [
    "geodetic2ecef",
    "ecef2geodetic",
    "ecef2enuv",
    "ecef2enu",
    "enu2uvw",
    "uvw2enu",
    "eci2geodetic",
    "geodetic2eci",
    "enu2ecef",
]

if typing.TYPE_CHECKING:
    from numpy import ndarray


def geodetic2ecef(
    lat: "ndarray", lon: "ndarray", alt: "ndarray", ell: Ellipsoid = None, deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    point transformation from Geodetic of specified ellipsoid (default WGS-84) to ECEF

    Parameters
    ----------

    lat : "ndarray"
           target geodetic latitude
    lon : "ndarray"
           target geodetic longitude
    h : "ndarray"
         target altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Returns
    -------

    ECEF (Earth centered, Earth fixed)  x,y,z

    x : "ndarray"
        target x ECEF coordinate (meters)
    y : "ndarray"
        target y ECEF coordinate (meters)
    z : "ndarray"
        target z ECEF coordinate (meters)
    """
    lat, ell = sanitize(lat, ell, deg)
    if deg:
        lon = radians(lon)

    # radius of curvature of the prime vertical section
    N = ell.semimajor_axis ** 2 / sqrt(ell.semimajor_axis ** 2 * cos(lat) ** 2 + ell.semiminor_axis ** 2 * sin(lat) ** 2)
    # Compute cartesian (geocentric) coordinates given  (curvilinear) geodetic
    # coordinates.
    x = (N + alt) * cos(lat) * cos(lon)
    y = (N + alt) * cos(lat) * sin(lon)
    z = (N * (ell.semiminor_axis / ell.semimajor_axis) ** 2 + alt) * sin(lat)

    return x, y, z


def ecef2geodetic(
    x: "ndarray", y: "ndarray", z: "ndarray", ell: Ellipsoid = None, deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    if vectorize is not None:
        fun = vectorize(ecef2geodetic_point)
        lat, lon, alt = fun(x, y, z, ell, deg)
        return lat[()], lon[()], alt[()]
    else:
        return ecef2geodetic_point(x, y, z, ell, deg)


def ecef2geodetic_point(x: float, y: float, z: float, ell: Ellipsoid = None, deg: bool = True) -> typing.Tuple[float, float, float]:
    """
    convert ECEF (meters) to geodetic coordinates

    Parameters
    ----------
    x : float
        target x ECEF coordinate (meters)
    y : float
        target y ECEF coordinate (meters)
    z : float
        target z ECEF coordinate (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    lat : float
           target geodetic latitude
    lon : float
           target geodetic longitude
    alt : float
         target altitude above geodetic ellipsoid (meters)

    based on:
    You, Rey-Jer. (2000). Transformation of Cartesian to Geodetic Coordinates without Iterations.
    Journal of Surveying Engineering. doi: 10.1061/(ASCE)0733-9453
    """
    if ell is None:
        ell = Ellipsoid()

    r = sqrt(x ** 2 + y ** 2 + z ** 2)

    E = sqrt(ell.semimajor_axis ** 2 - ell.semiminor_axis ** 2)

    # eqn. 4a
    u = sqrt(0.5 * (r ** 2 - E ** 2) + 0.5 * sqrt((r ** 2 - E ** 2) ** 2 + 4 * E ** 2 * z ** 2))

    Q = hypot(x, y)

    huE = hypot(u, E)

    # eqn. 4b
    try:
        Beta = atan(huE / u * z / hypot(x, y))
    except ZeroDivisionError:
        if z >= 0:
            Beta = pi / 2
        else:
            Beta = -pi / 2

    # eqn. 13
    eps = ((ell.semiminor_axis * u - ell.semimajor_axis * huE + E ** 2) * sin(Beta)) / (
        ell.semimajor_axis * huE * 1 / cos(Beta) - E ** 2 * cos(Beta)
    )

    Beta += eps
    # %% final output
    lat = atan(ell.semimajor_axis / ell.semiminor_axis * tan(Beta))

    lon = atan2(y, x)

    # eqn. 7
    alt = hypot(z - ell.semiminor_axis * sin(Beta), Q - ell.semimajor_axis * cos(Beta))

    # inside ellipsoid?
    inside = x ** 2 / ell.semimajor_axis ** 2 + y ** 2 / ell.semimajor_axis ** 2 + z ** 2 / ell.semiminor_axis ** 2 < 1
    if inside:
        alt = -alt

    if deg:
        lat = degrees(lat)
        lon = degrees(lon)

    return lat, lon, alt


def ecef2enuv(
    u: "ndarray", v: "ndarray", w: "ndarray", lat0: "ndarray", lon0: "ndarray", deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    VECTOR from observer to target  ECEF => ENU

    Parameters
    ----------
    u : "ndarray"
        target x ECEF coordinate (meters)
    v : "ndarray"
        target y ECEF coordinate (meters)
    w : "ndarray"
        target z ECEF coordinate (meters)
    lat0 : "ndarray"
           Observer geodetic latitude
    lon0 : "ndarray"
           Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    uEast : "ndarray"
        target east ENU coordinate (meters)
    vNorth : "ndarray"
        target north ENU coordinate (meters)
    wUp : "ndarray"
        target up ENU coordinate (meters)

    """
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lon0) * u + sin(lon0) * v
    uEast = -sin(lon0) * u + cos(lon0) * v
    wUp = cos(lat0) * t + sin(lat0) * w
    vNorth = -sin(lat0) * t + cos(lat0) * w

    return uEast, vNorth, wUp


def ecef2enu(
    x: "ndarray",
    y: "ndarray",
    z: "ndarray",
    lat0: "ndarray",
    lon0: "ndarray",
    h0: "ndarray",
    ell: Ellipsoid = None,
    deg: bool = True,
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    from observer to target, ECEF => ENU

    Parameters
    ----------
    x : "ndarray"
        target x ECEF coordinate (meters)
    y : "ndarray"
        target y ECEF coordinate (meters)
    z : "ndarray"
        target z ECEF coordinate (meters)
    lat0 : "ndarray"
           Observer geodetic latitude
    lon0 : "ndarray"
           Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Returns
    -------
    East : "ndarray"
        target east ENU coordinate (meters)
    North : "ndarray"
        target north ENU coordinate (meters)
    Up : "ndarray"
        target up ENU coordinate (meters)

    """
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)

    return uvw2enu(x - x0, y - y0, z - z0, lat0, lon0, deg=deg)


def enu2uvw(
    east: "ndarray", north: "ndarray", up: "ndarray", lat0: "ndarray", lon0: "ndarray", deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Parameters
    ----------

    e1 : "ndarray"
        target east ENU coordinate (meters)
    n1 : "ndarray"
        target north ENU coordinate (meters)
    u1 : "ndarray"
        target up ENU coordinate (meters)

    Results
    -------

    u : "ndarray"
    v : "ndarray"
    w : "ndarray"
    """

    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lat0) * up - sin(lat0) * north
    w = sin(lat0) * up + cos(lat0) * north

    u = cos(lon0) * t - sin(lon0) * east
    v = sin(lon0) * t + cos(lon0) * east

    return u, v, w


def uvw2enu(
    u: "ndarray", v: "ndarray", w: "ndarray", lat0: "ndarray", lon0: "ndarray", deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Parameters
    ----------

    u : "ndarray"
    v : "ndarray"
    w : "ndarray"


    Results
    -------

    East : "ndarray"
        target east ENU coordinate (meters)
    North : "ndarray"
        target north ENU coordinate (meters)
    Up : "ndarray"
        target up ENU coordinate (meters)
    """
    if deg:
        lat0 = radians(lat0)
        lon0 = radians(lon0)

    t = cos(lon0) * u + sin(lon0) * v
    East = -sin(lon0) * u + cos(lon0) * v
    Up = cos(lat0) * t + sin(lat0) * w
    North = -sin(lat0) * t + cos(lat0) * w

    return East, North, Up


def eci2geodetic(
    x: "ndarray", y: "ndarray", z: "ndarray", t: datetime, ell: Ellipsoid = None, deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    convert Earth Centered Internal ECI to geodetic coordinates

    J2000 time

    Parameters
    ----------
    x : "ndarray"
        ECI x-location [meters]
    y : "ndarray"
        ECI y-location [meters]
    z : "ndarray"
        ECI z-location [meters]
    t : datetime.datetime, "ndarray"
        UTC time
    ell : Ellipsoid (optional)
        planet ellipsoid model
    deg : bool (optional)
        if True, degrees. if False, radians

    Results
    -------
    lat : "ndarray"
          geodetic latitude
    lon : "ndarray"
          geodetic longitude
    alt : "ndarray"
          altitude above ellipsoid  (meters)

    eci2geodetic() a.k.a. eci2lla()
    """
    if eci2ecef is None:
        raise ImportError("pip install astropy")

    xecef, yecef, zecef = eci2ecef(x, y, z, t)

    return ecef2geodetic(xecef, yecef, zecef, ell, deg)


def geodetic2eci(
    lat: "ndarray", lon: "ndarray", alt: "ndarray", t: datetime, ell: Ellipsoid = None, deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    convert geodetic coordinates to Earth Centered Internal ECI

    J2000 frame

    Parameters
    ----------
    lat : "ndarray"
        geodetic latitude
    lon : "ndarray"
        geodetic longitude
    alt : "ndarray"
        altitude above ellipsoid  (meters)
    t : datetime.datetime, "ndarray"
        UTC time
    ell : Ellipsoid (optional)
        planet ellipsoid model
    deg : bool (optional)
        if True, degrees. if False, radians

    Results
    -------
    x : "ndarray"
        ECI x-location [meters]
    y : "ndarray"
        ECI y-location [meters]
    z : "ndarray"
        ECI z-location [meters]

    geodetic2eci() a.k.a lla2eci()
    """
    if ecef2eci is None:
        raise ImportError("pip install astropy")

    x, y, z = geodetic2ecef(lat, lon, alt, ell, deg)

    return ecef2eci(x, y, z, t)


def enu2ecef(
    e1: "ndarray",
    n1: "ndarray",
    u1: "ndarray",
    lat0: "ndarray",
    lon0: "ndarray",
    h0: "ndarray",
    ell: Ellipsoid = None,
    deg: bool = True,
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    ENU to ECEF

    Parameters
    ----------

    e1 : "ndarray"
        target east ENU coordinate (meters)
    n1 : "ndarray"
        target north ENU coordinate (meters)
    u1 : "ndarray"
        target up ENU coordinate (meters)
    lat0 : "ndarray"
        Observer geodetic latitude
    lon0 : "ndarray"
        Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Results
    -------
    x : "ndarray"
        target x ECEF coordinate (meters)
    y : "ndarray"
        target y ECEF coordinate (meters)
    z : "ndarray"
        target z ECEF coordinate (meters)
    """
    x0, y0, z0 = geodetic2ecef(lat0, lon0, h0, ell, deg=deg)
    dx, dy, dz = enu2uvw(e1, n1, u1, lat0, lon0, deg=deg)

    return x0 + dx, y0 + dy, z0 + dz



""" Transforms involving NED North East Down """

import typing

from .enu import geodetic2enu, aer2enu, enu2aer
from .ecef import ecef2geodetic, ecef2enuv, ecef2enu, enu2ecef
from .ellipsoid import Ellipsoid

if typing.TYPE_CHECKING:
    from numpy import ndarray


def aer2ned(
    az: "ndarray", elev: "ndarray", slantRange: "ndarray", deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    converts azimuth, elevation, range to target from observer to North, East, Down

    Parameters
    -----------

    az : "ndarray"
         azimuth
    elev : "ndarray"
         elevation
    slantRange : "ndarray"
         slant range [meters]
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------
    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)
    """
    e, n, u = aer2enu(az, elev, slantRange, deg=deg)

    return n, e, -u


def ned2aer(n: "ndarray", e: "ndarray", d: "ndarray", deg: bool = True) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    converts North, East, Down to azimuth, elevation, range

    Parameters
    ----------

    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)
    deg : bool, optional
        degrees input/output  (False: radians in/out)

    Results
    -------

    az : "ndarray"
         azimuth
    elev : "ndarray"
         elevation
    slantRange : "ndarray"
         slant range [meters]
    """
    return enu2aer(e, n, -d, deg=deg)


def ned2geodetic(
    n: "ndarray",
    e: "ndarray",
    d: "ndarray",
    lat0: "ndarray",
    lon0: "ndarray",
    h0: "ndarray",
    ell: Ellipsoid = None,
    deg: bool = True,
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Converts North, East, Down to target latitude, longitude, altitude

    Parameters
    ----------

    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)
    lat0 : "ndarray"
        Observer geodetic latitude
    lon0 : "ndarray"
        Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------

    lat : "ndarray"
        target geodetic latitude
    lon : "ndarray"
        target geodetic longitude
    h : "ndarray"
        target altitude above geodetic ellipsoid (meters)

    """
    x, y, z = enu2ecef(e, n, -d, lat0, lon0, h0, ell, deg=deg)

    return ecef2geodetic(x, y, z, ell, deg=deg)


def ned2ecef(
    n: "ndarray",
    e: "ndarray",
    d: "ndarray",
    lat0: "ndarray",
    lon0: "ndarray",
    h0: "ndarray",
    ell: Ellipsoid = None,
    deg: bool = True,
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    North, East, Down to target ECEF coordinates

    Parameters
    ----------

    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)
    lat0 : "ndarray"
        Observer geodetic latitude
    lon0 : "ndarray"
        Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------

    x : "ndarray"
        ECEF x coordinate (meters)
    y : "ndarray"
        ECEF y coordinate (meters)
    z : "ndarray"
        ECEF z coordinate (meters)
    """
    return enu2ecef(e, n, -d, lat0, lon0, h0, ell, deg=deg)


def ecef2ned(
    x: "ndarray",
    y: "ndarray",
    z: "ndarray",
    lat0: "ndarray",
    lon0: "ndarray",
    h0: "ndarray",
    ell: Ellipsoid = None,
    deg: bool = True,
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Convert ECEF x,y,z to North, East, Down

    Parameters
    ----------

    x : "ndarray"
        ECEF x coordinate (meters)
    y : "ndarray"
        ECEF y coordinate (meters)
    z : "ndarray"
        ECEF z coordinate (meters)
    lat0 : "ndarray"
        Observer geodetic latitude
    lon0 : "ndarray"
        Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------

    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)

    """
    e, n, u = ecef2enu(x, y, z, lat0, lon0, h0, ell, deg=deg)

    return n, e, -u


def geodetic2ned(
    lat: "ndarray",
    lon: "ndarray",
    h: "ndarray",
    lat0: "ndarray",
    lon0: "ndarray",
    h0: "ndarray",
    ell: Ellipsoid = None,
    deg: bool = True,
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    convert latitude, longitude, altitude of target to North, East, Down from observer

    Parameters
    ----------

    lat : "ndarray"
        target geodetic latitude
    lon : "ndarray"
        target geodetic longitude
    h : "ndarray"
        target altitude above geodetic ellipsoid (meters)
    lat0 : "ndarray"
        Observer geodetic latitude
    lon0 : "ndarray"
        Observer geodetic longitude
    h0 : "ndarray"
         observer altitude above geodetic ellipsoid (meters)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)


    Results
    -------

    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)
    """
    e, n, u = geodetic2enu(lat, lon, h, lat0, lon0, h0, ell, deg=deg)

    return n, e, -u


def ecef2nedv(
    x: "ndarray", y: "ndarray", z: "ndarray", lat0: "ndarray", lon0: "ndarray", deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    for VECTOR between two points

    Parameters
    ----------
    x : "ndarray"
        ECEF x coordinate (meters)
    y : "ndarray"
        ECEF y coordinate (meters)
    z : "ndarray"
        ECEF z coordinate (meters)
    lat0 : "ndarray"
        Observer geodetic latitude
    lon0 : "ndarray"
        Observer geodetic longitude
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------

    (Vector)

    n : "ndarray"
        North NED coordinate (meters)
    e : "ndarray"
        East NED coordinate (meters)
    d : "ndarray"
        Down NED coordinate (meters)
    """
    e, n, u = ecef2enuv(x, y, z, lat0, lon0, deg=deg)

    return n, e, -u



"""
converts right ascension, declination to azimuth, elevation and vice versa.
Normally do this via AstroPy.
These functions are fallbacks for those wihtout AstroPy.

Michael Hirsch implementation of algorithms from D. Vallado
"""

import typing
from datetime import datetime

try:
    from numpy import sin, cos, degrees, radians, arcsin as asin, arctan2 as atan2
except ImportError:
    from math import sin, cos, degrees, radians, asin, atan2

from .sidereal import datetime2sidereal

__all__ = ["azel2radec", "radec2azel"]

if typing.TYPE_CHECKING:
    from numpy import ndarray


def azel2radec(
    az_deg: "ndarray", el_deg: "ndarray", lat_deg: "ndarray", lon_deg: "ndarray", time: datetime, usevallado: bool = True
) -> typing.Tuple["ndarray", "ndarray"]:
    """
    converts azimuth, elevation to right ascension, declination

    Parameters
    ----------

    az_deg : "ndarray"
        azimuth (clockwise) to point [degrees]
    el_deg : "ndarray"
        elevation above horizon to point [degrees]
    lat_deg : "ndarray"
        observer WGS84 latitude [degrees]
    lon_deg : "ndarray"
        observer WGS84 longitude [degrees]
    time : datetime.datetime
        time of observation

    Results
    -------

    ra_deg : "ndarray"
        right ascension to target [degrees]
    dec_deg : "ndarray"
        declination of target [degrees]

    from D.Vallado Fundamentals of Astrodynamics and Applications
    p.258-259
    """

    if abs(lat_deg) > 90:
        raise ValueError("-90 <= lat <= 90")

    az = radians(az_deg)
    el = radians(el_deg)
    lat = radians(lat_deg)
    lon = radians(lon_deg)
    # %% Vallado "algorithm 28" p 268
    dec = asin(sin(el) * sin(lat) + cos(el) * cos(lat) * cos(az))

    lha = atan2(-(sin(az) * cos(el)) / cos(dec), (sin(el) - sin(lat) * sin(dec)) / (cos(dec) * cos(lat)))

    lst = datetime2sidereal(time, lon)  # lon, ra in RADIANS

    """ by definition right ascension [0, 360) degrees """
    return degrees(lst - lha) % 360, degrees(dec)


def radec2azel(
    ra_deg: "ndarray", dec_deg: "ndarray", lat_deg: "ndarray", lon_deg: "ndarray", time: datetime, usevallado: bool = True
) -> typing.Tuple["ndarray", "ndarray"]:
    """
    converts right ascension, declination to azimuth, elevation

    Parameters
    ----------

    ra_deg : "ndarray"
        right ascension to target [degrees]
    dec_deg : "ndarray"
        declination to target [degrees]
    lat_deg : "ndarray"
        observer WGS84 latitude [degrees]
    lon_deg : "ndarray"
        observer WGS84 longitude [degrees]
    time : datetime.datetime
        time of observation

    Results
    -------

    az_deg : "ndarray"
        azimuth clockwise from north to point [degrees]
    el_deg : "ndarray"
        elevation above horizon to point [degrees]


    from D. Vallado "Fundamentals of Astrodynamics and Applications "
       4th Edition Ch. 4.4 pg. 266-268
    """
    if abs(lat_deg) > 90:
        raise ValueError("-90 <= lat <= 90")

    ra = radians(ra_deg)
    dec = radians(dec_deg)
    lat = radians(lat_deg)
    lon = radians(lon_deg)

    lst = datetime2sidereal(time, lon)  # RADIANS
    # %% Eq. 4-11 p. 267 LOCAL HOUR ANGLE
    lha = lst - ra
    # %% #Eq. 4-12 p. 267
    el = asin(sin(lat) * sin(dec) + cos(lat) * cos(dec) * cos(lha))
    # %% combine Eq. 4-13 and 4-14 p. 268
    az = atan2(-sin(lha) * cos(dec) / cos(el), (sin(dec) - sin(el) * sin(lat)) / (cos(el) * cos(lat)))

    return degrees(az) % 360.0, degrees(el)



""" Line of sight intersection of space observer to ellipsoid """

import typing

try:
    from numpy import pi, nan, sqrt, vectorize
except ImportError:
    from math import pi, nan, sqrt

    vectorize = None

from .aer import aer2enu
from .ecef import enu2uvw, geodetic2ecef, ecef2geodetic
from .ellipsoid import Ellipsoid

__all__ = ["lookAtSpheroid"]

if typing.TYPE_CHECKING:
    from numpy import ndarray


def lookAtSpheroid(
    lat0: "ndarray", lon0: "ndarray", h0: "ndarray", az: "ndarray", tilt: "ndarray", ell: Ellipsoid = None, deg: bool = True
) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Calculates line-of-sight intersection with Earth (or other ellipsoid) surface from above surface / orbit

    Parameters
    ----------

    lat0 : "ndarray"
           observer geodetic latitude
    lon0 : "ndarray"
           observer geodetic longitude
    h0 : "ndarray"
        observer altitude (meters)  Must be non-negative since this function doesn't consider terrain
    az : "ndarray"
        azimuth angle of line-of-sight, clockwise from North
    tilt : "ndarray"
        tilt angle of line-of-sight with respect to local vertical (nadir = 0)
    ell : Ellipsoid, optional
          reference ellipsoid
    deg : bool, optional
          degrees input/output  (False: radians in/out)

    Results
    -------

    lat : "ndarray"
           geodetic latitude where the line-of-sight intersects with the Earth ellipsoid
    lon : "ndarray"
           geodetic longitude where the line-of-sight intersects with the Earth ellipsoid
    d : "ndarray"
        slant range (meters) from starting point to intersect point

    Values will be NaN if the line of sight does not intersect.

    Algorithm based on https://medium.com/@stephenhartzell/satellite-line-of-sight-intersection-with-earth-d786b4a6a9b6 Stephen Hartzell
    """
    if vectorize is not None:
        fun = vectorize(lookAtSpheroid_point)
        lat, lon, d = fun(lat0, lon0, h0, az, tilt, ell, deg)
        return lat[()], lon[()], d[()]
    else:
        return lookAtSpheroid_point(lat0, lon0, h0, az, tilt, ell, deg)


def lookAtSpheroid_point(
    lat0: float, lon0: float, h0: float, az: float, tilt: float, ell: Ellipsoid = None, deg: bool = True
) -> typing.Tuple[float, float, float]:

    if h0 < 0:
        raise ValueError("Intersection calculation requires altitude  [0, Infinity)")

    if ell is None:
        ell = Ellipsoid()

    a = ell.semimajor_axis
    b = ell.semimajor_axis
    c = ell.semiminor_axis

    el = tilt - 90.0 if deg else tilt - pi / 2

    e, n, u = aer2enu(az, el, srange=1.0, deg=deg)  # fixed 1 km slant range
    u, v, w = enu2uvw(e, n, u, lat0, lon0, deg=deg)
    x, y, z = geodetic2ecef(lat0, lon0, h0, deg=deg)

    value = -(a ** 2) * b ** 2 * w * z - a ** 2 * c ** 2 * v * y - b ** 2 * c ** 2 * u * x
    radical = (
        a ** 2 * b ** 2 * w ** 2
        + a ** 2 * c ** 2 * v ** 2
        - a ** 2 * v ** 2 * z ** 2
        + 2 * a ** 2 * v * w * y * z
        - a ** 2 * w ** 2 * y ** 2
        + b ** 2 * c ** 2 * u ** 2
        - b ** 2 * u ** 2 * z ** 2
        + 2 * b ** 2 * u * w * x * z
        - b ** 2 * w ** 2 * x ** 2
        - c ** 2 * u ** 2 * y ** 2
        + 2 * c ** 2 * u * v * x * y
        - c ** 2 * v ** 2 * x ** 2
    )

    magnitude = a ** 2 * b ** 2 * w ** 2 + a ** 2 * c ** 2 * v ** 2 + b ** 2 * c ** 2 * u ** 2

    # %%   Return nan if radical < 0 or d < 0 because LOS vector does not point towards Earth
    if radical > 0:
        d = (value - a * b * c * sqrt(radical)) / magnitude
    else:
        d = nan

    if d < 0:
        d = nan
    # %% cartesian to ellipsodal
    lat, lon, _ = ecef2geodetic(x + d * u, y + d * v, z + d * w, deg=deg)

    return lat, lon, d



""" transforms involving ECI earth-centered inertial """

from datetime import datetime
import typing

try:
    from astropy.coordinates import GCRS, ITRS, EarthLocation, CartesianRepresentation
    from astropy.time import Time
    import astropy.units as u
except ImportError:
    Time = None

__all__ = ["eci2ecef", "ecef2eci"]

if typing.TYPE_CHECKING:
    from numpy import ndarray


def eci2ecef(x: "ndarray", y: "ndarray", z: "ndarray", time: datetime) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Observer => Point  ECI  =>  ECEF

    J2000 frame

    Parameters
    ----------
    x : "ndarray"
        ECI x-location [meters]
    y : "ndarray"
        ECI y-location [meters]
    z : "ndarray"
        ECI z-location [meters]
    time : datetime.datetime
        time of obsevation (UTC)

    Results
    -------
    x_ecef : "ndarray"
        x ECEF coordinate
    y_ecef : "ndarray"
        y ECEF coordinate
    z_ecef : "ndarray"
        z ECEF coordinate
    """

    if Time is None:
        raise ImportError("pip install astropy")

    gcrs = GCRS(CartesianRepresentation(x * u.m, y * u.m, z * u.m), obstime=time)
    itrs = gcrs.transform_to(ITRS(obstime=time))

    return itrs.x.value, itrs.y.value, itrs.z.value


def ecef2eci(x: "ndarray", y: "ndarray", z: "ndarray", time: datetime) -> typing.Tuple["ndarray", "ndarray", "ndarray"]:
    """
    Point => Point   ECEF => ECI

    J2000 frame

    Parameters
    ----------

    x : "ndarray"
        target x ECEF coordinate
    y : "ndarray"
        target y ECEF coordinate
    z : "ndarray"
        target z ECEF coordinate
    time : datetime.datetime
        time of observation

    Results
    -------
    x_eci : "ndarray"
        x ECI coordinate
    y_eci : "ndarray"
        y ECI coordinate
    z_eci : "ndarray"
        z ECI coordinate
    """

    if Time is None:
        raise ImportError("pip install astropy")

    itrs = ITRS(CartesianRepresentation(x * u.m, y * u.m, z * u.m), obstime=time)
    gcrs = itrs.transform_to(GCRS(obstime=time))
    ecef = EarthLocation(*gcrs.cartesian.xyz)

    return ecef.x.value, ecef.y.value, ecef.z.value



# Copyright (c) 2014-2018 Michael Hirsch, Ph.D.
""" manipulations of sidereal time """
from math import pi
from datetime import datetime

from .timeconv import str2dt

try:
    from astropy.time import Time
    import astropy.units as u
    from astropy.coordinates import Longitude
except ImportError:
    Time = None

"""
The "usevallado" datetime to julian runs 4 times faster than astropy.
However, AstroPy is more accurate.
"""

__all__ = ["datetime2sidereal", "juliandate", "julian2sidereal"]


def datetime2sidereal(time: datetime, lon_radians: float, usevallado: bool = True) -> float:
    """
    Convert ``datetime`` to sidereal time

    from D. Vallado "Fundamentals of Astrodynamics and Applications"


    time : datetime.datetime
        time to convert
    lon_radians : float
        longitude (radians)
    usevallado : bool, optional
        use vallado instead of AstroPy (default is Vallado)

    Results
    -------

    tsr : float
        Sidereal time
    """
    if isinstance(time, (tuple, list)):
        return [datetime2sidereal(t, lon_radians) for t in time]

    usevallado = usevallado or Time is None
    if usevallado:
        jd = juliandate(str2dt(time))
        # %% Greenwich Sidereal time RADIANS
        gst = julian2sidereal(jd)
        # %% Algorithm 15 p. 188 rotate GST to LOCAL SIDEREAL TIME
        tsr = gst + lon_radians
    else:
        tsr = Time(time).sidereal_time(kind="apparent", longitude=Longitude(lon_radians, unit=u.radian)).radian

    return tsr


def juliandate(time: datetime) -> float:
    """
    Python datetime to Julian time

    from D.Vallado Fundamentals of Astrodynamics and Applications p.187
     and J. Meeus Astronomical Algorithms 1991 Eqn. 7.1 pg. 61

    Parameters
    ----------

    time : datetime.datetime
        time to convert

    Results
    -------

    jd : float
        Julian date
    """
    if isinstance(time, (tuple, list)):
        return list(map(juliandate, time))

    if time.month < 3:
        year = time.year - 1
        month = time.month + 12
    else:
        year = time.year
        month = time.month

    A = int(year / 100.0)
    B = 2 - A + int(A / 4.0)
    C = ((time.second / 60.0 + time.minute) / 60.0 + time.hour) / 24.0

    return int(365.25 * (year + 4716)) + int(30.6001 * (month + 1)) + time.day + B - 1524.5 + C


def julian2sidereal(Jdate: float) -> float:
    """
    Convert Julian time to sidereal time

    D. Vallado Ed. 4

    Parameters
    ----------

    Jdate: float
        Julian centuries from J2000.0

    Results
    -------

    tsr : float
        Sidereal time
    """
    if isinstance(Jdate, (tuple, list)):
        return list(map(julian2sidereal, Jdate))

    # %% Vallado Eq. 3-42 p. 184, Seidelmann 3.311-1
    tUT1 = (Jdate - 2451545.0) / 36525.0

    # Eqn. 3-47 p. 188
    gmst_sec = 67310.54841 + (876600 * 3600 + 8640184.812866) * tUT1 + 0.093104 * tUT1 ** 2 - 6.2e-6 * tUT1 ** 3

    # 1/86400 and %(2*pi) implied by units of radians
    return gmst_sec * (2 * pi) / 86400.0 % (2 * pi)
